/*
 * D-LITe : Copyright (c) 2011 Philipe Valembois - Sylvain Cherrier
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * 
 *
 */

#include "contiki.h"
#include "contiki-lib.h"
#include "contiki-net.h"

#include <string.h>

#include "coap.h"

#define DEBUG DEBUG_PRINT
#include "net/uip-debug.h"

#define WITH_BUTTON 1
#if WITH_BUTTON
#include "dev/button-sensor.h"
#endif
#include "dev/leds.h"
//#define ANNOUNCE_COUNT 3
//#define ANNOUNCE_INTERVAL 1 /* tous les COAP_WAIT secondes */

static struct uip_udp_conn *server_conn;

void notify();
void fst(const char *message);
static unsigned char notify_obs = 0;
static char *notify_message = NULL;

PROCESS(coap_server_process, "COAP server process");
AUTOSTART_PROCESSES(&coap_server_process);
/*---------------------------------------------------------------------------*/

void
send_reply(uip_ipaddr_t *addr, unsigned short port, unsigned short size)
{
  uip_ipaddr_copy(&server_conn->ripaddr, addr);
  server_conn->rport = port;

  uip_udp_packet_send(server_conn, uip_appdata, size);

  /* Restore server connection to allow data from any node */
  
  uip_create_unspecified(&server_conn->ripaddr);
  server_conn->rport = 0;
  
}

static void
udp_handler(void)
{
  if (uip_poll()) {
    notify();
  }
  else if(uip_newdata()) {
/*    
    ((char *)uip_appdata)[uip_datalen()] = 0;
    PRINTF("Server received: '%s' from ", (char *)uip_appdata);
    PRINT6ADDR(&UIP_IP_BUF->srcipaddr);
    PRINTF("\n");

    uip_ipaddr_copy(&server_conn->ripaddr, &UIP_IP_BUF->srcipaddr);
    server_conn->rport = UIP_UDP_BUF->srcport;
    PRINTF("Responding with message: ");
    sprintf(uip_appdata, "Hello from the server! (%d)", ++seq_id);
    PRINTF("%s\n", (char *) uip_appdata);
*/

    PRINTF("Server received: packet from ");
    PRINT6ADDR(&UIP_IP_BUF->srcipaddr);
    PRINTF("\n");

    uip_ipaddr_copy(&server_conn->ripaddr, &UIP_IP_BUF->srcipaddr);
    server_conn->rport = UIP_UDP_BUF->srcport;

    coap_handler(server_conn);   
 
    /* Restore server connection to allow data from any node */
    uip_create_unspecified(&server_conn->ripaddr);
    server_conn->rport = 0;
  }
}
/*---------------------------------------------------------------------------*/
static void
print_local_addresses(void)
{
  int i;
  uint8_t state;

  PRINTF("Server IPv6 addresses: ");
  for(i = 0; i < UIP_DS6_ADDR_NB; i++) {
    state = uip_ds6_if.addr_list[i].state;
    if(uip_ds6_if.addr_list[i].isused &&
       (state == ADDR_TENTATIVE || state == ADDR_PREFERRED)) {
      PRINT6ADDR(&uip_ds6_if.addr_list[i].ipaddr);
      PRINTF("\n");
    }
  }
}
/*---------------------------------------------------------------------------*/
PROCESS_THREAD(coap_server_process, ev, data)
{
#if UIP_CONF_ROUTER
  uip_ipaddr_t ipaddr;
#endif /* UIP_CONF_ROUTER */

  PROCESS_BEGIN();
  PRINTF("UDP server started\n");

#if UIP_CONF_ROUTER
  uip_ip6addr(&ipaddr, 0xaaaa, 0, 0, 0, 0, 0, 0, 0);
  uip_ds6_set_addr_iid(&ipaddr, &uip_lladdr);
  uip_ds6_addr_add(&ipaddr, 0, ADDR_AUTOCONF);
#endif /* UIP_CONF_ROUTER */

  print_local_addresses();

  coap_init();

  server_conn = udp_new(NULL, 0, NULL);
  udp_bind(server_conn, UIP_HTONS(COAP_PORT));

#if WITH_BUTTON
  button_sensor.configure(SENSORS_ACTIVE, 1);
#endif
  while(1) {
    PROCESS_YIELD();
    if(ev == tcpip_event) {
      if (uip_udpconnection()) {
        udp_handler();
      }
    }
#if WITH_BUTTON
    else if(ev == sensors_event && data == &button_sensor) {
      fst("push");  //an action is like receiving a message - to be adapted for each kind of hardware...
      printf("Button pressed \n");
    }
#endif
    if (notify_message) {
      tcpip_poll_udp(server_conn);
    }
  }

  PROCESS_END();
}
/*---------------------------------------------------------------------------*/

#define NB_RULES 7
#define NB_OBSERVERS 5
#define WORD_SIZE 6

typedef struct {
  char original_state[WORD_SIZE];
  char input[WORD_SIZE];
  char destination_state[WORD_SIZE];
  char message[WORD_SIZE];
} rule;

static rule rules[NB_RULES];
static unsigned char rules_number = 0;
// these two lines need to be adapted for each kind of hardware...
static char hardware_description[] = "name=TelosB\nmsg=push\nstate=dark\nstate=light";  //specific message, describing our hardware
static char current_state[WORD_SIZE] = "dark";

typedef struct {
  uip_ipaddr_t ip;
  unsigned short port;
  char uri[WORD_SIZE];
} observer;
static observer observers[NB_OBSERVERS];
static unsigned char obs_number = 0;

void notify() {
  if (!notify_message)
    return;

  if (notify_obs >= obs_number) {
    PRINTF("Finished with message %s\n", notify_message);
    notify_message = NULL;
    return;
  }

  char *data = coap_make_post_request(CT_PLAIN_FORMAT, observers[notify_obs].uri);
  data = data + sprintf((char *)data, "input=%s", notify_message);

  uip_ipaddr_copy(&server_conn->ripaddr, &observers[notify_obs].ip);
  server_conn->rport = UIP_HTONS(observers[notify_obs].port);

  PRINTF("Sending message %s to ", notify_message);
  PRINT6ADDR(&observers[notify_obs].ip);
  PRINTF(":%u\n", observers[notify_obs].port);
  uip_udp_packet_send(server_conn, uip_appdata, data - ((char *) uip_appdata));

  /* Restore server connection to allow data from any node */
  
  uip_create_unspecified(&server_conn->ripaddr);
  server_conn->rport = 0;

  notify_obs++;
  if (notify_obs >= obs_number) {
    notify_message = NULL;
  }
}

void action() { 				//- to be adapted for each kind of hardware...
  if (!strcmp(current_state,"dark")) {
    leds_off(LEDS_BLUE);
  } else if (!strcmp(current_state,"light")) {
    leds_on(LEDS_BLUE);
  }
}

void fst(const char *message) {  //here we're trying to find what has to be done
  int i;
  for(i = 0; (i < rules_number) && (i < NB_RULES); i++) {
    //PRINTF("checking rule %d : %s vs %s = %d\n",i, message, rules[i].input, strcmp(message, rules[i].input));
    if (strcmp(message, rules[i].input) == 0) { //rule found
      //PRINTF("checking state : %s vs %s = %d\n",current_state, rules[i].original_state, strcmp(current_state, rules[i].original_state));
      if (strcmp(current_state, rules[i].original_state) == 0) { //and good state
        strncpy(current_state, rules[i].destination_state, WORD_SIZE); //new state
        action(); //changing state.... checking if a real action is to be done...
        PRINTF("Changing to state %s\n", current_state);
        notify_message = rules[i].message; // observers need to be warned !!
        notify_obs = 0;
        break; // rule has been found
      }
    }
  }
}

static void prepare_payload(char *payload, unsigned short size) {
  for(;size > 0; size--, payload++) {
    if (*payload == '&') {
      *payload = '\0';
    }
  }
  // At the end we put a \0
  *payload = '\0';
}

static const char *find_variable(const char *payload, unsigned short size, const char *variable) {
  const char *needle = variable;
  const char *post = payload;
  while (post < payload + size) {
    if (*needle != '\0') {
      if (*post != *needle) {
        // We search the next \0 (formerly an &)
        for(;post < payload + size && *post != '\0'; post++)
          ;
        needle = variable;
      }
      else {
        needle++;
      }
    }
    else {
      if (*post != '=') {
        for(;post < payload + size && *post != '\0'; post++)
          ;
        needle = variable;
      }
      else {
        post++;
        break;
      }
    }
    post++;
  }
  if (post < payload + size) {
    return post;
  }
  else {
    return NULL;
  }
}

static char *reply_dlite(unsigned char method, char *payload, unsigned short size) {
  char *data;
  if (method == METHOD_GET) {
    data = coap_make_base_reply(CT_PLAIN_FORMAT);
    data = data + sprintf(data, "%s", hardware_description);  //GET if for getting description of DLite capacities of this node...
    return data;
  }
  else if (method == METHOD_POST) {
    prepare_payload(payload, size);
    const char *input = find_variable(payload, size, "input");  // POST is where we receive a message from other nodes
    if (!input) {
      data = coap_make_error_reply(ERROR_400);
      data = data + sprintf(data, "no input");
      return data;
    }
    // Here is where FST acts
    fst(input);
    data = coap_make_base_reply(CT_PLAIN_FORMAT);
    data = data + sprintf(data, "state=%s", current_state);
    return data;
  }
  else if (method == METHOD_PUT) {
    prepare_payload(payload, size);
    const char *order = find_variable(payload, size, "order");   // PUT is used for configuration (rules, initial states, observers...)
    if (!order) {
      data = coap_make_error_reply(ERROR_400);
      data = data + sprintf(data, "no order");
      return data;
    }
    if (!strcmp(order, "rule")) {
      if (rules_number >= NB_RULES) { // Too many rules
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "too many rules");
        return data;
      }
      const char *rule;
      rule = find_variable(payload, size, "state");
      if (!rule) {
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "no state");
        return data;
      }
      strncpy(rules[rules_number].original_state, rule, WORD_SIZE);
      rule = find_variable(payload, size, "msg");
      if (!rule) {
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "no msg");
        return data;
      }
      strncpy(rules[rules_number].input, rule, WORD_SIZE);
      rule = find_variable(payload, size, "Nstate");
      if (!rule) {
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "no Nstate");
        return data;
      }
      strncpy(rules[rules_number].destination_state, rule, WORD_SIZE);
      rule = find_variable(payload, size, "Smsg");
      if (!rule) {
        rule = "";
      }
      strncpy(rules[rules_number].message, rule, WORD_SIZE);
      data = coap_make_base_reply(CT_PLAIN_FORMAT);
      data = data + sprintf(data, "rule %d added", rules_number);
      rules_number++;
      return data;
    }
    else if (!strcmp(order, "init")) {
      const char *rule = find_variable(payload, size, "state");
      if (!rule) {
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "no state");
        return data;
      }
      strncpy(current_state, rule, WORD_SIZE);
      action();
      data = coap_make_base_reply(CT_PLAIN_FORMAT);
      data = data + sprintf(data, "initial state is %s", current_state);
      return data;
    }
    else if (!strcmp(order, "link")) {
      if (obs_number >= NB_OBSERVERS) { // Too many observers
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "too many observers");
        return data;
      }
      const char *uri = find_variable(payload, size, "uri");
      if (!uri) {
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "no URI");
        return data;
      }
      observers[obs_number].port = COAP_PORT;
      observers[obs_number].uri[0] = '\0';
      if (!uiplib_ipaddrconv(uri, &observers[obs_number].ip)) {
        data = coap_make_error_reply(ERROR_400);
        data = data + sprintf(data, "bad URI");
        return data;
      }
      if (uri[0] == '[') {
        // If there is a [, there is a ] and maybe something else that is a port or a uri
        for(; *uri != '\0' && *uri != ']'; uri++)
          ;
        if (*uri == '\0') {
          // Unfinished [
          data = coap_make_error_reply(ERROR_400);
          data = data + sprintf(data, "bad URI");
          return data;
        }
        uri++;
        if (*uri == ':') {
          uri++;
          observers[obs_number].port = 0;
          while (*uri > '0' && *uri < '9') {
            observers[obs_number].port = observers[obs_number].port*10 + (*uri - '0');
            uri++;
          }
        }
        if (*uri == '/') {
          uri++;
          strncpy(observers[obs_number].uri, uri, WORD_SIZE);
        }
        else {
          data = coap_make_error_reply(ERROR_400);
          data = data + sprintf(data, "bad URI");
          return data;
        }
      }
      data = coap_make_base_reply(CT_PLAIN_FORMAT);
      data = data + sprintf(data, "linked %d", obs_number);
      obs_number++;
      return data;
    }
    else {
      data = coap_make_error_reply(ERROR_400);
      data = data + sprintf(data, "bad order");
      return data;
    }
  }
  else if (method == METHOD_DELETE) {
    current_state[0] = '\0';
    rules_number = 0;
    obs_number = 0;
    notify_message = NULL;
    data = coap_make_error_reply(ERROR_205);
    data = data + sprintf(data, "deleted");
    return data;
  }
  else {
    data = coap_make_error_reply(ERROR_405);
    return data;
  }
}

#define URL_DLITE "dlite"
#define URL_WK_R ".well-known/r"
#define URL_WK_CORE ".well-known/core"

#define WELLKNOWN_R "</" URL_DLITE ">;n=\"DLite rule\""

static char *get_wk_r(unsigned char method, char *payload, unsigned short size) {
  char *data;
  if (method != METHOD_GET) {
      data = coap_make_error_reply(ERROR_405);
      return data;
  }
  data = coap_make_base_reply(CT_LINK_FORMAT);
  strcpy((char *) data, WELLKNOWN_R);
  return data + sizeof(WELLKNOWN_R) - 1;
}

static const observer_funcs services[] = { {URL_DLITE, reply_dlite}, 
                                           {URL_WK_R, get_wk_r}, 
                                           {URL_WK_CORE, get_wk_r} };

const observer_funcs obs_not_found = { NULL, coap_get_not_found };

const observer_funcs *get_from_url(char *url, unsigned short size) {
    int i;
    for(i=0; i<size; i++) {
        if (url[i] == '?')
            size = i + 1;
    }

    for(i = 0; i < sizeof(services) / sizeof(observer_funcs); i++) {
      if (strncmp(url, services[i].url, size) == 0) {
        return &services[i];
    }
    }
    return &obs_not_found;
}

