#include "contiki-lib.h"
#include "contiki-net.h"
#include <string.h>
#include "coap.h"

#define DEBUG DEBUG_PRINT
#include "net/uip-debug.h"

#define ANNOUNCE_URL "well-known/ann"

typedef struct {
  unsigned char vto;
  unsigned char code;
  unsigned short tid;
} coap_header;

#define COAP_HDR_VER(hdr) ((hdr->vto >> 6) & 0x3)
#define COAP_HDR_TYPE(hdr) ((hdr->vto >> 4) & 0x3)
#define COAP_HDR_OPTCOUNT(hdr) (hdr->vto & 0xf)
#define COAP_HDR_VTO(ver, type, opt) ((ver << 6) | ((type & 0x3) << 4) | (opt & 0xf))

#define COAP_OHDR_OFFSET(ohdr) ((*ohdr >> 4) & 0xf)
#define COAP_OHDR_SIZE(ohdr) (*ohdr & 0xf)
#define COAP_OHDR_OS(offset, size) (((offset) << 4) | ((size) & 0xf))

typedef struct {
  void *ptr;
  unsigned short size;
  union {
    char *str;
    int num;
  } value;
} option;

#define OPTID_URI_PATH 0
#define OPTTYPE_URI_PATH string
#define OPTID_SUB_LIFETIME 1
#define OPTTYPE_SUB_LIFETIME mbint
#define NUM_OPTID 2

#define PASTE(a, b) a ## b
#define PASTEU(a, b) a ## _ ## b

#define ID(opt) PASTE(OPTID_ , opt)
#define OPT(opt) PASTE(OPT_ , opt)
#define TYPE(opt) PASTE(OPTTYPE_ , opt)

#define MAP(opt) { OPT(opt) , ID(opt) } 

#define DO(meth, type, ...) PASTEU(meth, type)(__VA_ARGS__)
#define OPTION_PARSE(opt) DO(parse, TYPE(opt), &option_table[ID(opt)])
#define OPTION_GET(opt) (DO(get, TYPE(opt), &option_table[ID(opt)]))
#define OPTION_SIZE(opt) (option_table[ID(opt)].size)
#define OPTION_RESET(opt) option_table[ID(opt)].ptr = NULL

#define get_string(optent) ((optent)->value.str)
#define parse_string(optent) ((optent)->value.str = (char *) (optent)->ptr)

#define get_mbint(optent) ((optent)->value.num)
static char parse_mbint(option *optent) {
  if (optent->ptr == NULL)
    return 0;
  switch (optent->size) {
    case 1:
      optent->value.num = *((unsigned char *) optent->ptr);
      break;
    case 2:
      optent->value.num = UIP_HTONS(*((unsigned short *) optent->ptr));
      break;
    case 4:
      optent->value.num = UIP_HTONL(*((unsigned int *) optent->ptr));
      break;
    default:
      return 0;

  }
  return 1;
}

// A absolument trier par numero d'option
static const unsigned short option_maptable[][2] = { MAP(URI_PATH), MAP(SUB_LIFETIME) };

static option option_table[NUM_OPTID];

static coap_header *coap_hdr;

static unsigned short last_tid;

option *find_option(unsigned short optnum) {
  unsigned short i;

  for(i = 0; i < sizeof(option_maptable) / 2*sizeof(char); i++) {
    if (option_maptable[i][0] == optnum)
      return option_table + option_maptable[i][1];
  }
  return NULL;
}

void coap_handler(struct uip_udp_conn *conn) {
  static short lastOpt;
  unsigned short i;
  unsigned short size = 0;
  char *ohdr;
  option *opt;

//PRINTF("ver %u typ %u code %u tid %u\n", COAP_HDR_VER(coap_hdr), COAP_HDR_TYPE(coap_hdr), coap_hdr->code, coap_hdr->tid);
  
  if (COAP_HDR_VER(coap_hdr) != COAP_VERSION)
    return;

  if (COAP_HDR_TYPE(coap_hdr) == TYPE_ACK || COAP_HDR_TYPE(coap_hdr) == TYPE_RST) {
    return;
  }

  char confirmable = (COAP_HDR_TYPE(coap_hdr) == TYPE_CON);
  if (uip_is_addr_mcast(&UIP_IP_BUF->destipaddr)) {
    size = coap_make_reset(coap_hdr->tid) - ((char *) coap_hdr);
  }
  else if (COAP_HDR_TYPE(coap_hdr) == TYPE_ACK || COAP_HDR_TYPE(coap_hdr) == TYPE_RST) {
    size = coap_make_reset(coap_hdr->tid) - ((char *) coap_hdr);
  }
  else {
    if (coap_hdr->code == METHOD_GET && COAP_HDR_TYPE(coap_hdr) == TYPE_NON) {
      /* a GET without reply is non-sense */
      return;
    }
    else {
      OPTION_RESET(URI_PATH);
      OPTION_RESET(SUB_LIFETIME);
      ohdr = (char *) (coap_hdr + 1);
      lastOpt = 0;
      for(i = COAP_HDR_OPTCOUNT(coap_hdr); i > 0; i--) {
        lastOpt = lastOpt + COAP_OHDR_OFFSET(ohdr);
        opt = find_option(lastOpt);
        if (opt != NULL) {
          if (COAP_OHDR_SIZE(ohdr) == 0xf) {
            opt->size = 15 + *((char *) (ohdr + 1));
            opt->ptr = ((char *) (ohdr + 1)) + 1;
          }
          else {
            opt->size = COAP_OHDR_SIZE(ohdr);
            opt->ptr = (ohdr + 1);
          }
        }
        ohdr = ((char *) opt->ptr) + opt->size;
      }
      if (!OPTION_PARSE(URI_PATH)) {
        // Without URI we can't do nothing
        return;
      }
      const observer_funcs *service = get_from_url(OPTION_GET(URI_PATH), OPTION_SIZE(URI_PATH));
      /* ATTENTION !! ICI ON VA FAIRE UN GROS HACK !! */
      /* On utilise le meme buffer pour lire la payload et pour écrire notre réponse : ca permet de limiter la conso mémoire */
      /* Si on écrit trop loin on va écraser notre donnée recue mais bon on va suppose qu'on est sages et qu'on écrira juste comme il faut */
      /* Si il n'y a pas d'URL dans la requête ou si elle a une taille de 1 caractere, on sera possiblement mal (surtout avec un CT sur 2 octets ce que nous n'avons pas la) */
      char *payload = (char *) ohdr;
      char *data = (char *) coap_hdr;
      size = UIP_HTONS(UIP_UDP_BUF->udplen) - (payload - ((char *) UIP_UDP_BUF));
      size = service->reply(coap_hdr->code, payload, size) - (char *) coap_hdr;
    }
  }
  
  if (size && confirmable) {
PRINTF("Replying size %d ver %u typ %u code %u tid %u\n", size, COAP_HDR_VER(coap_hdr), COAP_HDR_TYPE(coap_hdr), coap_hdr->code, coap_hdr->tid);
    uip_udp_packet_send(conn, uip_appdata, size);
  }
 
}

char *coap_make_reset() {
  coap_hdr->vto = COAP_HDR_VTO(COAP_VERSION, TYPE_RST, 0);
  coap_hdr->code = 0;
  // coap_hdr->tid = coap_hdr->tid // NO-OP : it's a reply
  
  return (char *) (coap_hdr + 1);
}

char *coap_get_not_found(unsigned char method, char *payload, unsigned short size) {
  if (!uip_is_addr_mcast(&UIP_IP_BUF->destipaddr)) {
    coap_header *hdr = (coap_header *) coap_hdr;
    
    hdr->vto = COAP_HDR_VTO(COAP_VERSION, TYPE_ACK, 0);
    hdr->code = ERROR_404;
    // ((coap_header *)data)->tid = ((coap_header *)data)->tid // NO-OP : it's a reply
  
    return (char *) (hdr + 1);
  }
  else {
    return (char *) coap_hdr;
  }
}

char *coap_make_base_reply(unsigned char content_type) {
  coap_hdr->vto = COAP_HDR_VTO(COAP_VERSION, TYPE_ACK, 1);
  coap_hdr->code = ERROR_200;
  // coap_hdr->tid = coap_hdr->tid // NO-OP : it's a reply
  char *ohdr = (char *)(coap_hdr + 1);
  *(ohdr++) = COAP_OHDR_OS(OPT_CONTENT_TYPE, 1);
  *(ohdr++) = content_type;
  return ohdr;
}

char *coap_make_error_reply(unsigned char error_code) {
  coap_hdr->vto = COAP_HDR_VTO(COAP_VERSION, TYPE_ACK, 0);
  coap_hdr->code = error_code;
  // coap_hdr->tid = coap_hdr->tid // NO-OP : it's a reply
  return (char *)(coap_hdr + 1);
}

char *coap_make_post_request(unsigned char content_type, char *url) {
  unsigned char delta = 0;
  coap_hdr->vto = COAP_HDR_VTO(COAP_VERSION, TYPE_NON, 1);
  coap_hdr->code = METHOD_POST;
  last_tid++;
  coap_hdr->tid = UIP_HTONS(last_tid);
  unsigned short ln = strlen(url);
  char *ohdr = (char *)(coap_hdr + 1);
  if (content_type != 0) {
    *(ohdr++) = COAP_OHDR_OS(OPT_CONTENT_TYPE-delta, 1);
    delta = OPT_CONTENT_TYPE;
    *(ohdr++) = content_type;
  }
  if (ln == 0) {
    return ohdr;
  }
  if (ln < 15) {
    *ohdr = COAP_OHDR_OS(OPT_URI_PATH-delta, ln);
  }
  else {
    *(ohdr++) = COAP_OHDR_OS(OPT_URI_PATH-delta, 15);
    *ohdr = ln - 15;
  }
  
  ohdr++;
  memcpy(ohdr, url, ln);
  ohdr = ohdr + ln;

  return ohdr;
}

void coap_init() {
  last_tid = random_rand();
  coap_hdr = (coap_header *) &uip_buf[UIP_IPUDPH_LEN + UIP_LLH_LEN];
}

/*char *coap_make_announce() {
  coap_hdr->vto = COAP_HDR_VTO(COAP_VERSION, TYPE_NON, 1);
  coap_hdr->code = METHOD_POST;
  last_tid++;
  coap_hdr->tid = UIP_HTONS(last_tid);
  char *ohdr = (char *)(coap_hdr + 1);

  if (sizeof(ANNOUNCE_URL) <= 15) {
    *ohdr = COAP_OHDR_OS(OPT_URI_PATH, sizeof(ANNOUNCE_URL)-1);
  }
  else {
    *(ohdr++) = COAP_OHDR_OS(OPT_URI_PATH, 15);
    *ohdr = sizeof(ANNOUNCE_URL) - 15;
  }

  ohdr++;
  memcpy(ohdr, ANNOUNCE_URL, sizeof(ANNOUNCE_URL) - 1);
  ohdr = ohdr + sizeof(ANNOUNCE_URL) - 1;
  
  return ohdr;
}*/
